﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Ab5Edit
{
    public partial class frmMain : Form
    {
        private const int TILE_SIZE = 16;
        private const int TILE_PIXEL = TILE_SIZE / 8;
        private const int TTILE_SIZE = TILE_SIZE + 4;
        private const int TSELECT_SIZE = TTILE_SIZE - 1;
        private const int LSELECT_SIZE = TILE_SIZE - 1;

        private const byte MAX_WIDTH = 60;
        private const byte MAX_HEIGHT = 48;

        private const int FIRST_WALL = 71;
        private const int LAST_WALL = FIRST_WALL + 20;
        private const int COMPLEX_WALLS = FIRST_WALL + 11;

        private const byte TILE_KEY = 49;
        private const byte TILE_AMMO = 50;
        private const byte TILE_10CREDITS = 51;
        private const byte TILE_100CREDITS = 52;
        private const byte TILE_FIRSTAID = 53;
        private const byte TILE_EXTRALIFE = 54;
        private const byte TILE_VDOOR = 149;
        private const byte TILE_HDOOR = 150;

        private int[] gameNormalTileArray = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
                                             32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,5,56,54,5,59,5,5,5,63,
                                             64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
                                             96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
                                             128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150};
        private int[] gameHatcheryTileArray = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
                                               32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,5,56,54,6,59,6,6,6,63,
                                               64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
                                               96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
                                               128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,144,145,146,147,148,149,150};
        private int[] gameWallArray = { 0, 5, 7, 1, 8, 2, 10, 11, 6, 9, 3, 13, 4, 14, 12, 15 };
        private byte[] complexWallArray = { 0x03, 0x0C, 0x05, 0x0A, 0x0F };

        private Bitmap[] tileset;
        private byte numTiles;

        private byte levelWidth;
        private byte levelHeight;
        private byte levelFlags;
        private byte levelObjectives;
        private byte levelTimer;
        private byte levelStartX;
        private byte levelStartY;
        private byte levelFinishX;
        private byte levelFinishY;
        private ushort levelBTileOffset;
        private byte levelBTileType;
        private ushort levelDTileOffset;
        private byte levelDTileType;
        private byte levelEnemy1;
        private byte levelEnemy2;
        private byte levelBoss;

        private int selectedT, selectedX, selectedY = 0;
        private bool dragflag = false;

        private byte[,] tileData = new byte[60, 48];

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            loadTileset();
            clearLevel();
        }

        private void loadTileset()
        {
            // load tileset to array of Bitmaps
            BinaryReader tileFile = new BinaryReader(File.Open("tileset.dat", FileMode.Open));
            numTiles = tileFile.ReadByte();
            tileset = new Bitmap[numTiles];
            for (int t = 0; t < numTiles; t++)
            {
                tileset[t] = new Bitmap(TILE_SIZE, TILE_SIZE);
                using (Graphics gfx = Graphics.FromImage(tileset[t]))
                {
                    for (int b = 0; b < 8; b++)
                    {
                        byte mask = 0x80;
                        byte tilebyte = tileFile.ReadByte();
                        for (int i = 0; i < 8; i++)
                        {
                            int result = tilebyte & mask;
                            Rectangle rect = new Rectangle(i * TILE_PIXEL, b * TILE_PIXEL, TILE_PIXEL, TILE_PIXEL);
                            if (result == 0) gfx.FillRectangle(new SolidBrush(Color.FromArgb(173, 190, 174)), rect);
                            else gfx.FillRectangle(Brushes.Black, rect);
                            mask = (byte)(mask >> 1);
                        }
                    }
                }
            }
            tileFile.Close();
        }

        private void clearLevel()
        {
            // clear level header
            levelWidth = MAX_WIDTH;
            levelHeight = MAX_HEIGHT;
            levelFlags = 0;
            levelObjectives = 0;
            levelTimer = 60;
            levelStartX = 0;
            levelStartY = 0;
            levelFinishX = 0;
            levelFinishY = 0;
            levelBTileOffset = 0;
            levelBTileType = 0;
            levelDTileOffset = 0;
            levelDTileType = 0;
            levelEnemy1 = 0;
            levelEnemy2 = 0;
            levelBoss = 0;
            // clear tileData
            for (int x = 0; x < 60; x++)
            {
                for (int y = 0; y < 48; y++)
                {
                    tileData[x, y] = 5;
                }
            }
        }

        private void picTileset_Paint(object sender, PaintEventArgs e)
        {
            // fill background
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(0, 0, picTileset.Width, picTileset.Height));
            // draw all tileset tiles
            for (int t = 0; t < numTiles; t++)
            {
                e.Graphics.DrawImage(tileset[t], new Rectangle(((t % 32) * TTILE_SIZE) + 2, ((t / 32) * TTILE_SIZE) + 2, tileset[t].Width, tileset[t].Height));
            }
            // draw selection box
            e.Graphics.DrawRectangle(new Pen(Color.Orange, 1.0f), new Rectangle((selectedT % 32) * TTILE_SIZE, (selectedT / 32) * TTILE_SIZE, TSELECT_SIZE, TSELECT_SIZE));
            e.Graphics.DrawRectangle(new Pen(Color.Orange, 1.0f), new Rectangle(((selectedT % 32) * TTILE_SIZE) + 1, ((selectedT / 32) * TTILE_SIZE) + 1, TSELECT_SIZE - 2, TSELECT_SIZE - 2));
        }

        private void picLevel_Paint(object sender, PaintEventArgs e)
        {
            int keys = 0, doors = 0, ammo = 0, credits = 0, firstaids = 0, lives = 0;
            // fill background
            e.Graphics.FillRectangle(Brushes.Gray, new Rectangle(0, 0, picLevel.Width, picLevel.Height));
            // draw all level tiles
            for (int y = 0; y < levelHeight; y++)
            {
                for (int x = 0; x < levelWidth; x++)
                {
                    if (checkGameView.Checked == false)
                        // draw edit view
                        e.Graphics.DrawImage(tileset[tileData[x, y]], new Rectangle(x * TILE_SIZE, y * TILE_SIZE, TILE_SIZE, TILE_SIZE));
                    else
                    {
                        // draw game view
                        int dispTile;
                        if (!checkHatchery.Checked) dispTile = gameNormalTileArray[tileData[x, y]];
                        else dispTile = gameHatcheryTileArray[tileData[x, y]];
                        if ((dispTile >= FIRST_WALL) && (dispTile <= LAST_WALL))
                        {
                            bool[] wallArray = new bool[8];
                            wallArray[0] = checkWall(x, y - 1);
                            wallArray[1] = checkWall(x - 1, y);
                            wallArray[2] = checkWall(x + 1, y);
                            wallArray[3] = checkWall(x, y + 1);
                            wallArray[4] = checkWall(x - 1, y - 1);
                            wallArray[5] = checkWall(x + 1, y - 1);
                            wallArray[6] = checkWall(x - 1, y + 1);
                            wallArray[7] = checkWall(x + 1, y + 1);
                            byte wallsAdj = 0;
                            byte wallDiag = 0;
                            byte mask = 0x08;
                            for (int i = 0; i < 4; i++)
                            {
                                if (wallArray[i])
                                    wallsAdj = (byte)(wallsAdj | mask);
                                mask = (byte)(mask >> 1);
                            }
                            mask = 0x08;
                            for (int i = 4; i < 8; i++)
                            {
                                if (wallArray[i])
                                    wallDiag = (byte)(wallDiag | mask);
                                mask = (byte)(mask >> 1);
                            }
                            dispTile += gameWallArray[wallsAdj];
                            if (dispTile >= COMPLEX_WALLS)
                            {
                                wallDiag = (byte)(wallDiag ^ 0xFF);
                                wallDiag = (byte)(wallDiag & complexWallArray[dispTile - COMPLEX_WALLS]);
                                if (wallDiag == 0)
                                    dispTile += 5;
                            }
                        }
                        e.Graphics.DrawImage(tileset[dispTile], new Rectangle(x * TILE_SIZE, y * TILE_SIZE, TILE_SIZE, TILE_SIZE));
                    }
                    // check tile and add to stats if necessary
                    switch (tileData[x, y])
                    {
                        case TILE_KEY:
                            keys++;
                            break;
                        case TILE_VDOOR:
                        case TILE_HDOOR:
                            doors++;
                            break;
                        case TILE_AMMO:
                            ammo++;
                            break;
                        case TILE_10CREDITS:
                            credits += 10;
                            break;
                        case TILE_100CREDITS:
                            credits += 100;
                            break;
                        case TILE_FIRSTAID:
                            firstaids++;
                            break;
                        case TILE_EXTRALIFE:
                            lives++;
                            break;
                    }
                }
            }
            // draw selection box
            e.Graphics.DrawRectangle(new Pen(Color.Orange, 1.0f), new Rectangle(selectedX * TILE_SIZE, selectedY * TILE_SIZE, LSELECT_SIZE, LSELECT_SIZE));
            e.Graphics.DrawRectangle(new Pen(Color.Orange, 1.0f), new Rectangle((selectedX * TILE_SIZE) + 1, (selectedY * TILE_SIZE) + 1, LSELECT_SIZE - 2, LSELECT_SIZE - 2));
            // update stats bos
            textKeys.Text = keys.ToString();
            textDoors.Text = doors.ToString();
            textAmmo.Text = ammo.ToString();
            textCredits.Text = credits.ToString();
            textFirstAids.Text = firstaids.ToString();
            textLives.Text = lives.ToString();
        }

        private bool checkWall(int x, int y)
        {
            if ((x >= 0) && (x < levelWidth) && (y >= 0) && (y < levelHeight))
            {
                if ((tileData[x, y] >= FIRST_WALL) && (tileData[x, y] <= LAST_WALL)) return true;
                else return false;
            }
            else return false;
        }

        private void picTileset_MouseClick(object sender, MouseEventArgs e)
        {
            int newT = ((e.Y / TTILE_SIZE) * 32) + (e.X / TTILE_SIZE);
            if (newT < numTiles)
            {
                selectedT = newT;
                picTileset.Refresh();
                updateStatusBar();
            }
        }

        private void picLevel_MouseDown(object sender, MouseEventArgs e)
        {
            dragflag = true;
            int x = e.X / TILE_SIZE;
            int y = e.Y / TILE_SIZE;
            if ((x < levelWidth) && (y < levelHeight) && (x >= 0) && (y >= 0))
                tileData[x, y] = (byte)(selectedT);
        }

        private void picLevel_MouseUp(object sender, MouseEventArgs e)
        {
            dragflag = false;
            picLevel.Refresh();
        }

        private void picLevel_MouseMove(object sender, MouseEventArgs e)
        {
            int x = e.X / TILE_SIZE;
            int y = e.Y / TILE_SIZE;
            if ((x < levelWidth) && (y < levelHeight) && (x >= 0) && (y >= 0))
            {
                if (dragflag)
                {
                    tileData[x, y] = (byte)(selectedT);
                }
                selectedX = x;
                selectedY = y;
                picLevel.Refresh();
            }
            updateStatusBar();
        }

        private void updateStatusBar()
        {
            statusBar.Text = "X = " + selectedX.ToString() + ", Y = " + selectedY.ToString() + ", Offset = ";
            int tempInt = (selectedY * levelWidth) + selectedX;
            statusBar.Text += tempInt.ToString();
            statusBar.Text += ", Level tile = " + tileData[selectedX, selectedY] + ", Brush tile = " + selectedT;
        }

        private void buttonWidthDec_Click(object sender, EventArgs e)
        {
            if (levelWidth > 12)
            {
                levelWidth--;
                if (selectedX>=levelWidth)
                    selectedX = levelWidth - 1;
                picLevel.Refresh();
            }
        }

        private void buttonWidthInc_Click(object sender, EventArgs e)
        {
            if (levelWidth < MAX_WIDTH)
            {
                levelWidth++;
                picLevel.Refresh();
            }
        }

        private void buttonHeightDec_Click(object sender, EventArgs e)
        {
            if (levelHeight > 8)
            {
                levelHeight--;
                if (selectedY >= levelHeight)
                    selectedY = levelHeight - 1;
                picLevel.Refresh();
            }
        }

        private void buttonHeightInc_Click(object sender, EventArgs e)
        {
            if (levelHeight < MAX_HEIGHT)
            {
                levelHeight++;
                picLevel.Refresh();
            }
        }

        private void buttonNew_Click(object sender, EventArgs e)
        {
            // clear level if "Yes" is clicked
            if (MessageBox.Show("Are you sure?", "New Level", MessageBoxButtons.YesNo) == DialogResult.Yes)
            {
                clearLevel();
                picLevel.Refresh();
            }
        }

        private void buttonOpen_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog openFileWindow = new OpenFileDialog())
            {
                openFileWindow.Filter = "AB5 Level Files (*.ab5)|*.ab5";
                if (openFileWindow.ShowDialog() == DialogResult.OK)
                {
                    // first, clear level data
                    clearLevel();
                    // open file
                    BinaryReader levelFile = new BinaryReader(File.Open(openFileWindow.FileName, FileMode.Open));
                    byte tempByte = 0;
                    string tempStr;
                    char tempChar;
                    // read mission text
                    textMissionText.Text = "";
                    while (tempByte != 255)
                    {
                        // display penrow & pencol
                        tempByte = levelFile.ReadByte();
                        tempStr = tempByte.ToString();
                        textMissionText.Text += tempStr + ";";
                        tempByte = levelFile.ReadByte();
                        tempStr = tempByte.ToString();
                        textMissionText.Text += tempStr + ";";
                        tempByte = 1;
                        // read the rest of the line
                        while ((tempByte != 0) && (tempByte != 255))
                        {
                            tempByte = levelFile.ReadByte();
                            if ((tempByte != 0) && (tempByte != 255))
                            {
                                tempChar = (char)tempByte;
                                textMissionText.Text += tempChar;
                            }
                            else
                            {
                                tempStr = tempByte.ToString();
                                textMissionText.Text += ";" + tempStr;
                                if (tempByte == 0)
                                    textMissionText.Text += Environment.NewLine;
                            }
                        }
                    }
                    // read rest of the header
                    levelWidth = levelFile.ReadByte();
                    levelHeight = levelFile.ReadByte();
                    levelFlags = levelFile.ReadByte();
                    levelObjectives = levelFile.ReadByte();
                    levelTimer = levelFile.ReadByte();
                    levelStartX = levelFile.ReadByte();
                    levelStartY = levelFile.ReadByte();
                    levelFinishX = levelFile.ReadByte();
                    levelFinishY = levelFile.ReadByte();
                    levelBTileOffset = levelFile.ReadUInt16();
                    levelBTileType = levelFile.ReadByte();
                    levelDTileOffset = levelFile.ReadUInt16();
                    levelDTileType = levelFile.ReadByte();
                    levelEnemy1 = levelFile.ReadByte();
                    levelEnemy2 = levelFile.ReadByte();
                    levelBoss = levelFile.ReadByte();
                    // read tile data
                    for (int y = 0; y < levelHeight; y++)
                        for (int x = 0; x < levelWidth; x++)
                            tileData[x, y] = levelFile.ReadByte();
                    // close file and refresh level display and header vars
                    levelFile.Close();
                    picLevel.Refresh();
                    loadHeaderVars();
                }
            }
        }

        private void buttonSave_Click(object sender, EventArgs e)
        {
            // save header vars
            if (saveHeaderVars())
            {
                using (SaveFileDialog saveFileWindow = new SaveFileDialog())
                {
                    saveFileWindow.Filter = "AB5 Level Files (*.ab5)|*.ab5";
                    if (saveFileWindow.ShowDialog() == DialogResult.OK)
                    {
                        // open file
                        BinaryWriter levelFile = new BinaryWriter(File.Open(saveFileWindow.FileName, FileMode.Create));
                        // parse the mission objective text
                        char[] textDelimiters = { '\n' };
                        string[] lines = textMissionText.Text.Split(textDelimiters);
                        for (int l = 0; l < lines.Length; l++)
                        {
                            char[] lineDelimiters = { ';' };
                            string[] lineBlocks = lines[l].Split(lineDelimiters);
                            if (lineBlocks.Length == 4)
                            {
                                // write penrow, pencol, text, line terminator (NULL or $FF)
                                levelFile.Write(Convert.ToByte(lineBlocks[0]));
                                levelFile.Write(Convert.ToByte(lineBlocks[1]));
                                char[] lineText = lineBlocks[2].ToCharArray();
                                for (int c = 0; c < lineText.Length; c++)
                                    levelFile.Write((byte)(lineText[c]));
                                levelFile.Write(Convert.ToByte(lineBlocks[3]));
                            }
                            else MessageBox.Show("Mission objective text format error, text data may not have saved correctly!", "Save Level");
                        }
                        // save rest of the header
                        levelFile.Write(levelWidth);
                        levelFile.Write(levelHeight);
                        levelFile.Write(levelFlags);
                        levelFile.Write(levelObjectives);
                        levelFile.Write(levelTimer);
                        levelFile.Write(levelStartX);
                        levelFile.Write(levelStartY);
                        levelFile.Write(levelFinishX);
                        levelFile.Write(levelFinishY);
                        levelFile.Write(levelBTileOffset);
                        levelFile.Write(levelBTileType);
                        levelFile.Write(levelDTileOffset);
                        levelFile.Write(levelDTileType);
                        levelFile.Write(levelEnemy1);
                        levelFile.Write(levelEnemy2);
                        levelFile.Write(levelBoss);
                        // write tile data
                        for (int y = 0; y < levelHeight; y++)
                            for (int x = 0; x < levelWidth; x++)
                                levelFile.Write(tileData[x, y]);
                        // close file
                        levelFile.Close();
                    }
                }
            }
        }

        private void loadHeaderVars()
        {
            // load flag check boxes
            if ((levelFlags & 0x01) == 0) checkPasscode.Checked = false;
            else checkPasscode.Checked = true;
            if ((levelFlags & 0x02) == 0) checkFireDoors.Checked = false;
            else checkFireDoors.Checked = true;
            if ((levelFlags & 0x04) == 0) checkPowerDomes.Checked = false;
            else checkPowerDomes.Checked = true;
            if ((levelFlags & 0x08) == 0) checkDark.Checked = false;
            else checkDark.Checked = true;
            if ((levelFlags & 0x10) == 0) checkHatchery.Checked = false;
            else checkHatchery.Checked = true;
            if ((levelFlags & 0x20) == 0) checkFinish.Checked = false;
            else checkFinish.Checked = true;
            // load text boxes with other header info
            textObjectives.Text = levelObjectives.ToString();
            textTimer.Text = levelTimer.ToString();
            textStartX.Text = levelStartX.ToString();
            textStartY.Text = levelStartY.ToString();
            textFinishX.Text = levelFinishX.ToString();
            textFinishY.Text = levelFinishY.ToString();
            textBTileOffset.Text = levelBTileOffset.ToString();
            textBTileType.Text = levelBTileType.ToString();
            textDTileOffset.Text = levelDTileOffset.ToString();
            textDTileType.Text = levelDTileType.ToString();
            int tempInt = levelEnemy1;
            tempInt &= 0x1F;
            textEnemy1Type.Text = tempInt.ToString();
            tempInt = levelEnemy1;
            tempInt &= 0xE0;
            tempInt = tempInt >> 5;
            textEnemy1Level.Text = tempInt.ToString();
            tempInt = levelEnemy2;
            tempInt &= 0x1F;
            textEnemy2Type.Text = tempInt.ToString();
            tempInt = levelEnemy2;
            tempInt &= 0xE0;
            tempInt = tempInt >> 5;
            textEnemy2Level.Text = tempInt.ToString();
            tempInt = levelBoss;
            tempInt &= 0x07;
            textBossType.Text = tempInt.ToString();
            tempInt = levelBoss;
            tempInt &= 0x38;
            tempInt = tempInt >> 3;
            textBossLevel.Text = tempInt.ToString();
        }

        private bool saveHeaderVars()
        {
            // save flag check boxes
            levelFlags = 0;
            if (checkPasscode.Checked) levelFlags |= 0x01;
            if (checkFireDoors.Checked) levelFlags |= 0x02;
            if (checkPowerDomes.Checked) levelFlags |= 0x04;
            if (checkDark.Checked) levelFlags |= 0x08;
            if (checkHatchery.Checked) levelFlags |= 0x10;
            if (checkFinish.Checked) levelFlags |= 0x20;
            // save other header info from text boxes
            try
            {
                levelObjectives = Convert.ToByte(textObjectives.Text);
                levelTimer = Convert.ToByte(textTimer.Text);
                levelStartX = Convert.ToByte(textStartX.Text);
                levelStartY = Convert.ToByte(textStartY.Text);
                levelFinishX = Convert.ToByte(textFinishX.Text);
                levelFinishY = Convert.ToByte(textFinishY.Text);
                levelBTileOffset = Convert.ToUInt16(textBTileOffset.Text);
                levelBTileType = Convert.ToByte(textBTileType.Text);
                levelDTileOffset = Convert.ToUInt16(textDTileOffset.Text);
                levelDTileType = Convert.ToByte(textDTileType.Text);
                byte tempByteMSB, tempByteLSB;
                tempByteLSB = Convert.ToByte(textEnemy1Type.Text);
                tempByteMSB = Convert.ToByte(textEnemy1Level.Text);
                tempByteMSB = (byte)(tempByteMSB << 5);
                levelEnemy1 = (byte)(tempByteMSB | tempByteLSB);
                tempByteLSB = Convert.ToByte(textEnemy2Type.Text);
                tempByteMSB = Convert.ToByte(textEnemy2Level.Text);
                tempByteMSB = (byte)(tempByteMSB << 5);
                levelEnemy2 = (byte)(tempByteMSB | tempByteLSB);
                tempByteLSB = Convert.ToByte(textBossType.Text);
                tempByteMSB = Convert.ToByte(textBossLevel.Text);
                tempByteMSB = (byte)(tempByteMSB << 3);
                levelBoss = (byte)(tempByteMSB | tempByteLSB);
            }
            catch (FormatException e)
            {
                MessageBox.Show("Format error in header!", "Save Level");
                return false;
            }
            catch (OverflowException e)
            {
                MessageBox.Show("Overflow error in header!", "Save Level");
                return false;
            }
            return true;
        }

        private void checkGameView_CheckedChanged(object sender, EventArgs e)
        {
            picLevel.Refresh();
        }

        private void checkHatchery_CheckedChanged(object sender, EventArgs e)
        {
            picLevel.Refresh();
        }
    }
}
